import json

from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.http import HttpResponse

from telegraphy import Gateway


@login_required
def get_token(request):
    user = request.user
    gateway = Gateway.from_settings(settings=settings.TELEGRAPHY_CONF)
    token = gateway.get_token(user=user.pk)
    response = {
        'auth_token': token,
    }
    return HttpResponse(json.dumps(response))



from functools import wraps


@wraps
def for_client(f):
    '''Method decroator to indicate that a view is by the client'''
    return f


@wraps
def for_webapp(f):
    '''Method decroator to indicate that a view is by the webapp'''
    return f



from datetime import datetime, timedelta
from .decorators import for_client, for_webapp
from telegraphy.utils import attr_or_item, import_class
import xmlrpclib
import uuid

try:
    import json
except ImportError:
    import simple_json as json


class NotRegistered(Exception):
    pass


class ConfigurationError(Exception):
    pass


class AuthToken(object):

    """docstring for AuthToken"""

    def __init__(self):
        self.creation_time = datetime.now()
        self.expiration = timedelta(seconds=5)  # TODO: Configure
        self.uuid = uuid.uuid3(uuid.NAMESPACE_DNS, str(self.creation_time))
        self.value = str(self.uuid)

    def is_valid(self):
        return (self.creation_time + self.expiration) >= datetime.now()

    def __str__(self):
        return str(self.value)


class GatewayProxy(object):

    """Gateway proxy is called to send events to send envets to the gateay.
    The initial implementation is based on XMLRPC from stdlib, but it can be tunned
    for better performance."""

    def send_event(event_name, event_data):
        """Sends to the configured gateway an event"""
        raise NotImplementedError()

    @classmethod
    def from_settings(cls, settings):
        engine = attr_or_item(settings, 'TELEGRAPHY_RPC_ENGINE')
        rpc_params = attr_or_item(settings, 'TELEGRAPHY_RPC_PARAMS')
        engine_class = import_class(engine)
        instance = engine_class(**rpc_params)
        return instance


class XMLRPCGatewayProxy(GatewayProxy):

    """XMLRPC gateway implementation"""

    def __init__(self, url):
        """Constructor shuld not be called directly but useing base class
        ``from_settings`` method to ensure consitency"""
        self.proxy = xmlrpclib.ServerProxy(url)

    def send_event(self, name, data):
        return self.proxy.send_event(name, data)


class BaseEvent(object):

    """Event base class"""

    name = None  # Name of event
    data = None  # Filled upon instantiation

    _settings = None

    def __init__(self, data=None, serialized_data=None):
        """Event instances are sent though Transports to Gateway using the send() method.
        Data must be json serializable"""
        if data is not None:
            assert serialized_data is None, "Serialized data and data must not be provided at once"
            self.data = self.serialize(data)
        elif serialized_data is not None:
            self.data = self.unserizlise(serialized_data)

    @classmethod
    def get_gateway_proxy(cls):
        # TODO: Configure
        if cls._settings is None:
            raise NotRegistered(
                "Event has not been yet registered on any gateway.")
        return XMLRPCGatewayProxy.from_settings(cls._settings)

    def send(self):
        '''Send the event to the gateway'''
        return self.get_gateway_proxy().send_event(self.name, self.data)

    def serialize(self, data):
        '''Data is a python object
        Executed in the sender'''
        return json.dumps(data)

    def unserizlise(self, data):
        '''Data is a string.
        Executed in the gateway'''
        return json.loads(data)

    def apply_filter(self, **filter_arguments):
        """Returns True y event match with filters, Fasle otherwise"""
        # TODO: Complete this code
        return

    _generic_event_registry = {}

    @classmethod
    def generic_event_factory(cls, event_name):
        """Creates a generic event based on its name. If a event class is not registered
        in the gateway, and TELEGRAPHY_SEND_UNREGISTERED is turned on, this method is
        called to build the class. Not it hass a class level cache."""
        if not event_name in cls._generic_event_registry:
            cls._generic_event_registry = type()


class Gateway(object):

    # Known events name - event class
    registry = {}

    @for_webapp
    def register(self, event_class):
        '''Register a new event class in the Gateway'''

        if event_class is BaseEvent:
            raise ConfigurationError("Cannot register Event base class. "
                                     "Must be a subclass")

        if not issubclass(event_class, BaseEvent):
            raise ConfigurationError("Events must be subclass of BaseEvent")

        if not event_class.name:
            raise ConfigurationError("Event %s has no name." % event_class)

        if event_class.name in self.registry:
            raise ConfigurationError(
                "%s has already been registered" % event_class.name)

        # Copy on event only those parametters that are relevant
        event_communication_settings = {
            'TELEGRAPHY_RPC_ENGINE': self.settings.TELEGRAPHY_RPC_ENGINE,
            'TELEGRAPHY_RPC_PARAMS': self.settings.TELEGRAPHY_RPC_PARAMS,
        }

        event_class._settings = event_communication_settings

        self.registry[event_class.name] = event_class

    auth_tokens = []

    @for_webapp
    def get_auth_token(self, uid):
        '''Called from Web Application'''
        token = AuthToken()
        self.auth_tokens.append(token)
        return str(token)

    def expire_auth_token(self, auth_token):
        pass

    def auth_token_cleanup(self):
        '''Called periodically'''
        pass

    def verify_auth_token(self, auth_token):
        for token in self.auth_tokens:
            if token.value == auth_token:
                return True
        raise Exception("INVALID")

    @for_client
    def on_connect(self, web_socket, auth_token, session_token):
        '''Called when a Client sends CONNECT request
        returns connection success/failure '''
        if not self.verify_auth_token(auth_token):
            web_socket.write(self.ERROR_AUTH)
            web_socket.close()
        else:
            pass
        self.expire_auth_token(auth_token)

    @for_webapp
    def autodiscover(self, initial_path=None):
        '''Finds event classes accross project and auto register them'''
        if not initial_path:
            pass

    # TODO Check if still needed
    _subscriptions = {}

    @property
    def subscriptions(self):
        '''Reurns a dict event names and client tokens'''
        return self._subscriptions

    # Number of seconds before exipring lost connection token
    timeout = 3

    @classmethod
    def from_settings(cls, settings):
        """Gateway process factory"""
        engine_path = attr_or_item(settings, 'TELEGRAPHY_ENGINE')
        engine_class = import_class(engine_path)
        instance = engine_class(settings)
        return instance

    @for_webapp
    def on_event(self, event_name, event_data):
        '''Called from webapp through proxy.
        Rehidrates the event'''
        event_class = self.registry.get(event_name)
        if not event_class:
            if attr_or_item(self.settings, 'TELEGRAPHY_SEND_UNREGISTERED'):
                event_class = BaseEvent.generic_event_factory(event_name)
            else:
                return False
        event_intance = event_class(serialized_data=event_data)
        self.publish_to_subscribers(event_intance)
        # TODO: Return something sane (defer?)
        return True

    def publish_to_subscribers(self, event_intance):
        for conn in self.connections:
            pass

    def run(self):
        """Executes gateway server"""
        raise NotImplementedError()

    def getPubSubUris(self):
        """Returns registered pubsub events"""
        return ['http://telegraphy.machinalis.com/events#']

    @property
    def rpc_uri(self):
        """WAMP URI for RPC calls"""
        return attr_or_item(self.settings, 'TELEGRAPHY_RPC_URI')

    @property
    def event_prefix(self):
        """WAMP PubSub prefix for events"""
        return attr_or_item(self.settings, 'TELEGRAPHY_EVENT_PREFIX')



from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
# from django.utils import autoreload #TODO: Make autoreload work
from telegraphy import Gateway
from telegraphy.utils import check_valid_settings
from optparse import make_option


class Command(BaseCommand):
    args = '<p_id p_id ...>'
    help = 'Run telegraphy socket server'

    option_list = (
        make_option('-r', '--reload',
                    default=False,
                    action='store_true',
                    help="Enables autoreload of server. DEBUG setting must be set to True."
                    ),
    ) + BaseCommand.option_list

    def handle(self, *args, **options):
        # TODO: Make this work with twisted
        #autoreload.main(self.inner_run, args, options)
        self.inner_run(*args, **options)

    def inner_run(self, *args, **options):
        from telegraphy.contrib.django_telegraphy import settings
        if not check_valid_settings(settings):
            raise ImproperlyConfigured("Missing telegraphy configuration in "
                                       "settings")

        gateway = Gateway.from_settings(settings)
        gateway.run()



"""Collection of utility functions not tight to any other module"""
import sys
import re
from importlib import import_module
from traceback import format_exc
from functools import wraps


def build_url_from_settings(settings):
    """Constructs web socket urls from settings"""
    # TODO: Configure in settings
    is_secure = settings.TELEGRAPHY_IS_SECURE
    proto = 'wss' if is_secure else 'ws'
    url_parts = [proto,
                 '://',
                 (settings.TELEGRAPHY_WS_HOST or 'localhost'),
                 ':',
                 str(settings.TELEGRAPHY_WS_PORT),
                 '/',
                 (settings.TELEGRAPHY_WS_URI or '')]
    return ''.join(url_parts)


def check_valid_settings(settings):
    """Validate sane settings"""
    return True


def extract_host_from_request(request):
    """Extracts host from possible Django request"""
    host = None
    if not request.is_secure():
        host = request.META['HTTP_HOST']
    else:
        host = request.META['HTTPS_HOST']
    if ':' in host:
        host = host.split(':')[0]
    return host


def attr_or_item(obj, name, default=None):
    """Helper for settings provided either as module constants or dict keys"""
    if hasattr(obj, name):
        return getattr(obj, name)
    try:
        return obj[name]
    except KeyError:
        if default is not None:
            return default
        raise


def import_class(path):
    """Imports a class based in a string path"""
    requested_module, requested_class = path.rsplit('.', 1)
    module = import_module(requested_module)
    requested_class = getattr(module, requested_class)
    return requested_class


def camelcase_to_undersocre(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def underscore_to_camelcase(value, first_cap=True):
    words = value.split('_')
    if not first_cap:
        words = words[:1] + map(lambda x: x.capitalize(), words[1:])
    else:
        words = map(lambda x: x.capitalize(), words)
    return ''.join(words)

def get_user(context):
    """Gets user id and username from context"""
    user = context['request'].user
    if user.is_authenticated():
        user_data = (user.pk, user.username)
    else:
        user_data = (None, None)
    return user_data


def show_traceback(f):
    """Decorator for immediate execption reporting on twisted deferreds."""
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            sys.stderr.write(e)
            sys.stderr.writelines(format_exc())
            raise e
    return wrapped


from django.conf import settings as django_settings

DEBUG = django_settings.DEBUG
# The autodiscover will search for modules named EVENTS_MODULE_NAME in each
# installed app, to register the existing events.
# If a file other than events.py wants to be used, this setting must be changed.
EVENTS_MODULE_NAME = 'events'

# Engine for Gateway
TELEGRAPHY_ENGINE = getattr(django_settings, 'TELEGRAPHY_ENGINE',
                            'telegraphy.gateway.txwamp.TxWAMPGateway')
# Web socket port
TELEGRAPHY_WS_PORT = getattr(django_settings, 'TELEGRAPHY_WS_PORT', 9000)
# Host, if none, it's dynamically taken from context
TELEGRAPHY_WS_HOST = getattr(django_settings, 'TELEGRAPHY_WS_HOST', None)
# Base URL for WS
TELEGRAPHY_WS_URI = getattr(django_settings, 'TELEGRAPHY_WS_URI', None)

# RPC class
TELEGRAPHY_RPC_ENGINE = getattr(django_settings, 'TELEGRAPHY_RPC_ENGINE',
                                'telegraphy.gateway.base.XMLRPCGatewayProxy')
# URL for events from webapp to gateway
TELEGRAPHY_RPC_PARAMS = getattr(django_settings, 'TELEGRAPHY_RPC_PARAMS',
                                {'url': 'http://localhost:4000'})

# Send unrgiestered events
TELEGRAPHY_SEND_UNREGISTERED = getattr(django_settings,
                                       'TELEGRAPHY_SEND_UNREGISTERED', True)


# Javascript URL
# Alternative sources in http://autobahn.ws/js/downloads/
AUTOBAHN_URL = getattr(django_settings, 'AUTOBAHN_URL', django_settings.STATIC_URL +
                       ('telegraphy/js/autobahn.js' if DEBUG
                        else 'telegraphy/js/autobahn.js')
                       )

TELEGRAPHY_IS_SECURE = getattr(django_settings, 'TELEGRAPHY_IS_SECURE', False)

# WAMP prefix constants
TELEGRAPHY_RPC_URI = getattr(django_settings, 'TELEGRAPHY_RPC_URI',
                             'http://telegraphy.machinalis.com/rpc#')

TELEGRAPHY_EVENT_PREFIX = getattr(django_settings, 'TELEGRAPHY_EVENT_PREFIX',
                                  'http://telegraphy.machinalis.com/events#')





class Client(object):
    def __init__(self, transport=None, gateway=None):
        """Generic Client"""
        self.sequence_id = None
        self.session_token = None
        self.subscriptions = []

    def on_open(self):
        pass

    def on_message(self, json_message):
        pass

    def on_close(self):
        pass

    _state = None

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        assert new_state in MEERKAT_PROTOCOL_STATES
        self._state = new_state


import sys
import os

from twisted.application import internet, service
from twisted.web import server, resource, wsgi, static
from twisted.python import threadpool
from twisted.internet import reactor

from django.core.management.base import BaseCommand
from optparse import make_option
# from django.utils import autoreload #TODO: Make autoreload work
#from telegraphy import Gateway

PORT = 8080


class RootResource(resource.Resource):

    def __init__(self, wsgi_resource):
        resource.Resource.__init__(self)
        self.wsgi_resource = wsgi_resource

    def getChild(self, path, request):
        path0 = request.prepath.pop(0)
        request.postpath.insert(0, path0)
        return self.wsgi_resource


class Command(BaseCommand):

    help_text = '''Twited based django server'''
    option_list = (
        make_option('-p', '--port', type=int, default=8000,
            help="Port to bind to"),

    ) + BaseCommand.option_list

    def handle(self, *args, **options):

        from django.core.handlers.wsgi import WSGIHandler

        def wsgi_resource():
            pool = threadpool.ThreadPool()
            pool.start()
            # Allow Ctrl-C to get you out cleanly:
            reactor.addSystemEventTrigger('after', 'shutdown', pool.stop)
            wsgi_resource = wsgi.WSGIResource(reactor, pool, WSGIHandler())
            return wsgi_resource


        # Twisted Application Framework setup:
        application = service.Application('twisted-django')

        # WSGI container for Django, combine it with twisted.web.Resource:
        # XXX this is the only 'ugly' part: see the 'getChild' method in twresource.Root
        wsgi_root = wsgi_resource()
        root = RootResource(wsgi_root)

        # Servce Django media files off of /media:
        staticrsrc = static.File(os.path.join(os.path.abspath("."), "mydjangosite/media"))
        root.putChild("media", staticrsrc)

        # Serve it up:
        main_site = server.Site(root)

        conn = reactor.listenTCP(options['port'], main_site)

        reactor.run()




"""Real-time, Model-based events"""

import datetime
import importlib
import inspect
import xmlrpclib

from telegraphy.contrib.django_telegraphy import settings
from django.core import serializers
from django.db.models.signals import post_save, post_delete


ISO8601_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

# This list keeps a possibly out-of-data registry of registered events.
# TODO: a method to update the list through the Gateway
_events = []


def get_registered_events():
    """Return the list of currently registered events."""
    # TODO: add an optional parameter to update the list before returning.
    # TODO: Such update shall be through the Gateway.
    return _events


class BaseEventModel(object):
    """
    Base class for events-generating Models.

    """
    OP_CREATE = 'create'
    OP_UPDATE = 'update'
    OP_DELETE = 'delete'
    # Class' meta attributes
    model = None
    fields = None
    exclude = None
    operations = (OP_CREATE, OP_UPDATE, OP_DELETE)
    name = None
    verbose_name = None

    def __init__(self):
        self.name = self.name or self.get_default_name()
        self.verbose_name = self.verbose_name or self.get_default_verbose_name()

        gateway_proxy_url = settings.TELEGRAPHY_RPC_PARAMS['url']
        self.gateway_proxy = xmlrpclib.Server(gateway_proxy_url,
                                              allow_none=True)

    def get_default_name(self):
        """
        The event's default name (if none is given) is the related model's app
        followed by a dot, followed by the model's name.

        """
        model = self.get_target_model()
        module = model.__module__.split('.')[-2]  # Miss the prefix and .models
        return '.'.join([module, model.__name__])

    def get_default_verbose_name(self):
        """
        The event's default name (if none is given) is the related model's app
        followed by a dot, followed by the model's name.

        """
        model = self.get_target_model()
        return model._meta.verbose_name

    def get_target_model(self):
        if type(self.__class__) == BaseEventModel:
            raise NotImplementedError("Use a specific Event class")
        if isinstance(self.model, str):
            raise NotImplementedError(
                "Feature not supported yet. model must be a class")
        return self.model

    def on_model_create(self, sender, instance, created, raw, **kwargs):
        if created and not raw:
            self.send_to_gateway(instance, self.OP_CREATE)

    def on_model_update(self, sender, instance, created, raw, **kwargs):
        if not created and not raw:
            self.send_to_gateway(instance, self.OP_UPDATE)

    def on_model_delete(self, sender, instance, **kwargs):
        self.send_to_gateway(instance, self.OP_DELETE)

    def register(self):
        """
        Connect to the Django signals following the configured operations.
        """
        global _event

        sender = self.get_target_model()

        if self.OP_CREATE in self.operations:
            post_save.connect(self.on_model_create, sender=sender)

        if self.OP_UPDATE in self.operations:
            post_save.connect(self.on_model_update, sender=sender)

        if self.OP_DELETE in self.operations:
            post_delete.connect(self.on_model_delete, sender=sender)

        if self.operations:
            _events.append(self)

    def send_to_gateway(self, instance, event_type):
        """
        Serialize the event, with the given 'event_type' and send to
        configured gateway.

        """
        # Timestamp formatted with a profile of ISO 8601
        # http://www.w3.org/TR/NOTE-datetime
        timestamp = datetime.datetime.utcnow()
        timestamp = timestamp.strftime(ISO8601_TIME_FORMAT)
        meta = {'event_type': event_type,
                'verbose_name': self.verbose_name,
                'timestamp': timestamp}

        if hasattr(instance, 'to_dict'):
            data = instance.to_dict()
        else:
            data = self.to_dict(instance)
        event = {'name': self.name,
                 'meta': meta,
                 'data': data}
        self.gateway_proxy.send_event(event)

    def to_dict(self, instance):
        """Return a dict representation of the model instance's fields."""

        excluded = self.exclude or []
        if not self.fields:
            fields = [f.name for f in instance._meta.fields
                      if f.name not in excluded]
        else:
            fields = self.fields

        return dict((field, getattr(instance, field, '')) for field in fields)


def autodiscover():
    """
    For all the installed apps in the Django project, register all the
    BaseEventModel subclasses. They are looked for in the 'events.py' files.

    """
    if not _events:
        events_module_name = settings.EVENTS_MODULE_NAME

        def is_event_model(module):
            return (inspect.isclass(module) and
                    issubclass(module, BaseEventModel) and
                    module != BaseEventModel)

        for app in settings.django_settings.INSTALLED_APPS:
            events_module = None
            try:
                app_events_module = '%s.%s' % (app, events_module_name)
                events_module = importlib.import_module(app_events_module)
            except ImportError:
                # No events_module in this app.
                pass
            event_classes = inspect.getmembers(events_module, is_event_model)
            for cname, EventClass in event_classes:
                print "Event classes detected in app %s: %s" % (app, cname)
                event = EventClass()
                event.register()



# encoding: utf-8

from setuptools import setup, find_packages

with open('requirements.txt', 'rb') as fp:
    requirements = [req for req in fp.read().split() if req]

# Filter does not work, workarround
packages = [p for p in find_packages() if not 'demo_project' in p]

setup(
    name='telegraphy',
    version='0.1.1.4',
    description=('Telegraphy - Real Time Events For Django.'),
    long_description=open('README.rst').read(),
    changelog='',
    license='Apache License 2.0',
    author='Nahuel Defossé',
    author_email='nahuel.defosse@gmail.com',
    url='http://telegraphy.machinalis.com',
    platforms=('Any'),
    install_requires = requirements,
    packages = packages,
    zip_safe = False,
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    #
    classifiers = ["License :: OSI Approved :: Apache Software License",
                   "Development Status :: 3 - Alpha",
                   "Environment :: Console",
                   "Framework :: Django",
                   "Framework :: Twisted",
                   "Intended Audience :: Developers",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Topic :: Internet",
                   "Topic :: Software Development :: Libraries"],
    keywords = 'django real-time-web websocket wamp'
)



"""Template tags for Django templates"""

from django import template

from django.core.exceptions import ImproperlyConfigured
from socket import error as SocketError
import errno
from telegraphy.contrib.django_telegraphy import settings
from telegraphy.gateway.base import GatewayProxy
from telegraphy.utils import (build_url_from_settings,
                              extract_host_from_request,
                              get_user)
from django.template.loader import render_to_string
try:
    import json
except ImportError:
    import simple_json as json



register = template.Library()


@register.simple_tag(takes_context=True)
def auth_token(context):
    """Communicates with running gateway"""
    try:
        # Create communication with running gateway
        proxy = GatewayProxy.from_settings(settings)
        # Get user from conext (ReuqestContext instance should have it)
        user_id, username = get_user(context)
        token = proxy.get_auth_token(user_id, username)
        return token
    except SocketError as e:
        if e.errno == errno.ECONNREFUSED:
            return "ERROR:NOGATEWAY"
    except Exception as e:
        return "ERROR:%s" % e


@register.simple_tag(takes_context=True)
def telegraphy_scripts(context):
    """Creates telegrphy javascript inclusions and settings"""

    registered_model_events = json.dumps([])

    context =  {
        'TELEGRAPHY_EVENT_PREFIX': settings.TELEGRAPHY_EVENT_PREFIX,
        'TELEGRAPHY_RPC_URI': settings.TELEGRAPHY_RPC_URI,
        'TELEGRAPHY_WS_URL': telegraphy_ws_url(context),
        'AUTOBAHN_URL': settings.AUTOBAHN_URL,
        'registered_model_events': registered_model_events,
    }

    return render_to_string('django_telegraphy/telegraphy_scripts.html',context)


@register.simple_tag(takes_context=True)
def telegraphy_ws_url(context):
    """Returns the gateway web socket URL"""
    host = settings.TELEGRAPHY_WS_HOST

    request = context.get('request')
    if not request:
        raise ImproperlyConfigured("Request is not present in context")
    # TODO: Check if this comparision is valid and/or sane
    host = extract_host_from_request(request)
    if settings.TELEGRAPHY_WS_HOST is not None:
        if settings.TELEGRAPHY_WS_HOST != host:
            raise ImproperlyConfigured(
                "TELEGRAPHY_WS_HOST and current host do not match!")
    return build_url_from_settings(settings)


@register.simple_tag(takes_context=True)
def telegraphy_event_prefix(context):
    # TODO: Generalize
    return settings.TELEGRAPHY_EVENT_PREFIX



from mock import patch, MagicMock

# from django.db.models.signals import post_save, post_delete
from django.utils.unittest import TestCase

from events import (BaseEventModel, post_save, post_delete,
                    get_registered_events, ISO8601_TIME_FORMAT)


class BaseModelEventBaseTestClass(TestCase):
    def setUp(self):
        patcher = patch.object(BaseEventModel, 'get_default_name')
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = patch.object(BaseEventModel, 'get_default_verbose_name')
        patcher.start()
        self.addCleanup(patcher.stop)
        self.event = BaseEventModel()


class BaseModelEventSendToGatewayTests(BaseModelEventBaseTestClass):
    def setUp(self):
        super(BaseModelEventSendToGatewayTests, self).setUp()
        self.event.name = 'TestEventModel'
        self.event.gateway_proxy = MagicMock()
        self.instance = MagicMock()
        self.instance.to_dict = MagicMock()


    def test_if_instance_has_to_dict_it_is_used(self):
        self.event.send_to_gateway(self.instance, None)
        self.instance.to_dict.assert_called_once_with()

    def test_instance_dont_have_to_dict_self_method_is_used(self):
        delattr(self.instance, 'to_dict')
        with patch.object(self.event, 'to_dict') as mock_serialize:
            self.event.send_to_gateway(self.instance, None)
            mock_serialize.assert_called_once_with(self.instance)

    def test_gateway_proxy_send_event_is_called(self):
        test_type = MagicMock()
        #from telegraphy.contrib.django_telegraphy.events import datetime
        with patch('telegraphy.contrib.django_telegraphy.events.datetime'
                ) as mock_time:
            self.event.send_to_gateway(self.instance, test_type)
            timestamp = mock_time.datetime.utcnow.return_value.strftime.return_value
            meta = {'event_type': test_type,
                    'verbose_name': self.event.verbose_name,
                    'timestamp': timestamp}
            expected_data =  {'name': self.event.name,
                              'meta': meta,
                              'data': self.instance.to_dict.return_value}
            self.event.gateway_proxy.send_event.assert_called_once_with(expected_data)

    def test_event_meta_timestamp_has_ISO_format(self):
        with patch('telegraphy.contrib.django_telegraphy.events.datetime'
                ) as mock_time:
            self.event.send_to_gateway(self.instance, None)
            event = self.event.gateway_proxy.send_event.call_args[0][0]
            timestamp_format = mock_time.datetime.utcnow.return_value.strftime
            time_fmt = timestamp_format.call_args[0][0]
            self.assertEqual(time_fmt, ISO8601_TIME_FORMAT)


class BaseModelEventToDictTests(BaseModelEventBaseTestClass):

    def setUp(self):
        super(BaseModelEventToDictTests, self).setUp()
        self.instance = MagicMock()
        self.instance.foo = 'bar'
        self.instance.chan = 99

    def test_fields_is_defined_then_only_they_appear_in_output(self):
        self.event.fields = ('foo',)
        data = self.event.to_dict(self.instance)
        self.assertDictEqual(data, {'foo': 'bar'})

    def test_fields_defined_not_in_instance_then_set_empty(self):
        self.event.fields = ('mor', )
        self.instance.mor = None
        delattr(self.instance, 'mor')
        data = self.event.to_dict(self.instance)
        self.assertEqual(data['mor'], '')

    def test_fields_not_defined_then_instance__meta_fields_in_output(self):
        self.event.fields = None
        field = MagicMock()
        field.name = 'mor'
        setattr(self.instance, field.name, 'xxx')
        self.instance._meta.fields = [field]
        data = self.event.to_dict(self.instance)
        self.assertEqual(data[field.name], 'xxx')

    def test_fields_not_defined_then_in_output_if_not_in_exclude(self):
        self.event.fields = None
        self.event.exclude = ('mor', )
        field = MagicMock()
        field.name = 'mor'
        setattr(self.instance, field.name, 'xxx')
        self.instance._meta.fields = [field]
        data = self.event.to_dict(self.instance)
        self.assertNotIn('mor', data)


class BaseModelEventRegisterTests(BaseModelEventBaseTestClass):

    def setUp(self):
        super(BaseModelEventRegisterTests, self).setUp()
        self.event.operations = ()

        patcher = patch.object(self.event, 'get_target_model')
        self.mock_get_model = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = patch.object(post_save, 'connect')
        self.mock_connect = patcher.start()
        self.addCleanup(patcher.stop)

    def test_get_target_model_is_called(self):
        self.event.register()
        self.mock_get_model.assert_called_once_with()

    def test_no_operations_listed_then_post_save_not_connected(self):
        self.event.operations = ()
        self.event.register()
        self.assertEqual(self.mock_connect.call_count, 0)

    def test_if_OP_CREATE_then_post_save_is_connected_to_on_model_create_(self):
        self.event.operations = (BaseEventModel.OP_CREATE, )
        self.event.register()
        self.mock_connect.assert_called_once_with(
            self.event.on_model_create, sender=self.mock_get_model.return_value)

    def test_if_OP_UPDATE_then_post_save_is_connected_to_on_model_update(self):
        self.event.operations = (BaseEventModel.OP_UPDATE, )
        self.event.register()
        self.mock_connect.assert_called_once_with(
            self.event.on_model_update, sender=self.mock_get_model.return_value)

    def test_if_OP_DELETE_then_post_save_is_connected_to_on_model_delete(self):
        self.event.operations = (BaseEventModel.OP_DELETE, )
        with patch.object(post_delete, 'connect') as mock_connect:
            self.event.register()
            mock_connect.assert_called_once_with(
            self.event.on_model_delete, sender=self.mock_get_model.return_value)

    def test_any_operation_then_event_added_to_module_events(self):
        self.event.operations = MagicMock()
        assert len(get_registered_events()) == 0
        self.event.operations = (BaseEventModel.OP_CREATE, )
        self.event.register()
        self.assertItemsEqual(get_registered_events(), [self.event])




from telegraphy.gateway import Gateway
from telegraphy.utils import build_url_from_settings
from twisted.internet import reactor  # , defer

from autobahn.websocket import listenWS
from autobahn.wamp import exportRpc, \
    WampServerFactory, \
    WampServerProtocol
from twisted.web import xmlrpc, server
from urlparse import urlparse
from telegraphy.utils import show_traceback


class TelegraphyConnection(WampServerProtocol):

    """
    Demonstrates creating a simple server with Autobahn WebSockets that
    responds to RPC calls.
    """

    @show_traceback
    def onSessionOpen(self):
        # .. and register them for RPC. that's it.
        print "On Session Open"

        self.registerForRpc(self, self.gateway.rpc_uri)
        self.registerForPubSub(self.gateway.event_prefix, prefixMatch=True)

    def connectionLost(self, reason):
        WampServerProtocol.connectionLost(self, reason)
        self.factory.removeConnection(self)

    def clientSubscriptions(self, uri, args, extra=None):
        pass

    @exportRpc("authenticate")
    def authenticate(self, auth_token, session_id):
        print "Authentication request", auth_token, session_id
        if self.gateway.verify_auth_token(auth_token):
            return "OK"
        # TODO: Return error

    @exportRpc('publish')
    def clientPublish(self, uri, args, extra=None):
        """RPC method. Publish event from client"""
        print uri, args, extra
        self.factory.dispatch(self.gateway.event_prefix + uri, args)


class GatewayWampServerFactory(WampServerFactory):

    def __init__(self, *args, **kwargs):
        self.gateway = kwargs.pop('gateway')
        WampServerFactory.__init__(self, *args, **kwargs)
        self.connected_clients = []

    def buildProtocol(self, addr):
        protocol = WampServerFactory.buildProtocol(self, addr)
        protocol.gateway = self.gateway
        self.connected_clients.append(protocol)
        return protocol

    def onClientSubscribed(self, proto, topicUri):
        print "Client", proto, "subscribed to", topicUri


    def removeConnection(self, proto):
        self.connected_clients.remove(proto)


class WebAppXMLRPCInterface(xmlrpc.XMLRPC):

    """Web Application interface"""

    def __init__(self, *args, **kwargs):
        self.gateway = kwargs.pop('gateway')
        xmlrpc.XMLRPC.__init__(self, allowNone=True, useDateTime=True)

    # All exposed methods must start with xmlrpc

    def xmlrpc_get_auth_token(self):
        """Generate auth token"""
        return self.gateway.get_auth_token()

    def xmlrpc_send_event(self, event):
        """Method called from the web app side to publish an event to clients"""
        return self.gateway.send(event)



class TxWAMPGateway(Gateway):

    """Twitsed implementation of Gateway over WAMP protocol"""

    def __init__(self, settings):
        # Save settings for later use through properties
        self.settings = settings

    _url = None

    @property
    def url(self):
        if not self._url:
            self._url = build_url_from_settings(self.settings)
        return self._url

    @property
    def rpc_url(self):
        return self.settings.TELEGRAPHY_RPC_PARAMS['url']

    @property
    def debug(self):
        # TODO: Check if WAMP debug must be dettached from this setting
        return self.settings.DEBUG

    @property
    def xmlrpc_port(self):
        return urlparse(self.rpc_url).port

    def run(self, start_reactor=True):
        # Look for event classes in the source tree
        self.autodiscover()
        # Create factory
        self.factory = GatewayWampServerFactory(self.url,
                                                debugWamp=self.debug,
                                                gateway=self
                                                )
        print self.factory
        self.factory.protocol = TelegraphyConnection
        self.factory.setProtocolOptions(allowHixie76=True)
        listenWS(self.factory)

        r = WebAppXMLRPCInterface(gateway=self)

        reactor.listenTCP(self.xmlrpc_port, server.Site(r))

        if start_reactor:
            reactor.run()

    def send(self, event):
        topicUri = self.event_prefix + event['name']
        print "Dispatching ", topicUri, event
        self.factory.dispatch(topicUri, event)
